/*
 * AIDE (Advanced Intrusion Detection Environment)
 *
 * Copyright (C) 1999-2002, 2010, 2020-2023 Rami Lehti, Pablo Virolainen,
 *               Richard van den Berg, Hannes von Haugwitz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _ERRORCODES_H_INCLUDED
#define  _ERRORCODES_H_INCLUDED

#define ERROR_WRITING_ERROR 14
#define INVALID_ARGUMENT_ERROR 15
#define UNIMPLEMENTED_FUNCTION_ERROR 16
#define INVALID_CONFIGURELINE_ERROR 17
#define IO_ERROR 18
#define VERSION_MISMATCH_ERROR 19
#define EXEC_ERROR 20
#define LOCK_ERROR 21
#define MEMORY_ALLOCATION_FAILURE 22
#define THREAD_ERROR 23
#define HASH_ALGO_ERROR 30
#define DATABASE_ERROR 24
#define SIGNAL_INTERRUPT_ERROR 25

#endif
